/*
   +----------------------------------------------------------------------+
   | PHP Version 5                                                        |
   +----------------------------------------------------------------------+
   | Copyright (c) 2010 PrimeBase Technologies GmbH, Germany              |
   +----------------------------------------------------------------------+
   | This source file is subject to version 3.01 of the PHP license,      |
   | that is bundled with this package in the file LICENSE, and is        |
   | available through the world-wide-web at the following url:           |
   | http://www.php.net/license/3_01.txt                                  |
   | If you did not receive a copy of the PHP license and are unable to   |
   | obtain it through the world-wide-web, please send a note to          |
   | license@php.net so we can mail you a copy immediately.               |
   +----------------------------------------------------------------------+
   | Authors: Barry Leslie <barry.leslie@Primebase.com>                   |
   +----------------------------------------------------------------------+

	PHP extension for the PrimeBase BLOB Management System (PBMS) daemon.
======================
zend_parse_parameters() Info because this always confuses me.

Here is the full list of type specifiers:

    * l _ long
    * d _ double
    * s _ string (with possible null bytes) and its length
    * b _ boolean
    * r _ resource, stored in zval*
    * a _ array, stored in zval*
    * o _ object (of any class), stored in zval*
    * O _ object (of class specified by class entry), stored in zval*
    * z _ the actual zval*

The following characters also have a meaning in the specifier string:

    * | _ indicates that the remaining parameters are optional_ The storage variables corresponding to these parameters should be initialized to default values by the extension, since they will not be touched by the parsing function if the parameters are not passed_
    * / _ the parsing function will call SEPARATE_ZVAL_IF_NOT_REF() on the parameter it follows, to provide a copy of the parameter, unless it's a reference_
    * ! _ the parameter it follows can be of specified type or NULL (only applies to a, o, O, r, and z)_ If NULL value is passed by the user, the storage pointer will be set to NULL_ 


======================
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <dlfcn.h>

#include "php.h"
#include "php_ini.h"

#include "ext/standard/info.h"
#include "php_pbms.h"
#if HAVE_PBMSLIB_H

#include "pbmslib.h"
#include "php_streams.h"
#ifdef ZTS
	#include "TSRM.h"
#endif

#define PBMS_PHP_VERSION "0.2.1"

ZEND_BEGIN_MODULE_GLOBALS(pbms)
	long debug_mode;
	long display_errors;
	long default_connection_id;
	long num_connections;
	long max_connections;
	long pool_size;
	long default_port;
	char *default_host;
	char *default_database;
ZEND_END_MODULE_GLOBALS(pbms)

#ifdef ZTS
# define PBMS_LS_D	zend_pbms_globals *pbms_globals
# define PBMS_LS_DC	, PBMS_LS_D
# define PBMS_LS_C	pbms_globals
# define PBMS_LS_CC , PBMS_LS_C
# define PBMS_G(v) (pbms_globals->v)
# define PBMS_LS_FETCH()	zend_pbms_globals *pbms_globals = ts_resource(pbms_globals_id)
#else
# define PBMS_LS_D
# define PBMS_LS_DC
# define PBMS_LS_C
# define PBMS_LS_CC
# define PBMS_G(v) (pbms_globals.v)
# define PBMS_LS_FETCH()
#endif


#ifndef TRANSFER_BUFFER_SIZE
#define TRANSFER_BUFFER_SIZE (10 * 1024)
#endif

#define PBMS_CONNECTION_NAME	"PrimeBase pbms connection"
#define PBMS_POOL_NAME			"PrimeBase pbms pool"
#define PBMS_MAX_HOST_SIZE		124	/* Maximum length of an IP address allowed.*/
#define PBMS_NAME_SIZE			64		
#define MAX_ERROR				256

#define	PBMS_BASE_ERROR			5000
#define PHP_PBMS_CONID_ERR	(PBMS_BASE_ERROR + 1)
#define PHP_PBMS_CALLBACK_ERR (PBMS_BASE_ERROR + 2)
#define PHP_PBMS_MEMORY_ERR (PBMS_BASE_ERROR + 3)

#define SAFE_STRING(s) ((s)?(s):"")


typedef unsigned char Bool;
#undef TRUE
#undef FALSE
#undef OK
#undef FAILED
#define TRUE	1
#define FALSE	0
#define OK		TRUE
#define FAILED	FALSE

typedef struct {
	PBMS	pbms;	
	
	char	host[PBMS_MAX_HOST_SIZE];
	long	port;
	char	database[PBMS_NAME_SIZE];	/* A list of open databases. */	
	
	Bool	display_error; 		/* Report the error immediatly.*/
	char	err_msg[MAX_ERROR+1];	/* Error text.*/
	int		err_no;
	
	Bool	in_use; 
	struct PBMSPool_tag *my_pool; 
} PBMSCon, *PBMSConPtr;

#define CLEAR_ERRORS(pbcon) {pbcon->err_no = 0;pbcon->err_msg[0] = 0;}

#ifndef MAX_PBMS_POOL_CONNECTIONS
#define MAX_PBMS_POOL_CONNECTIONS 100
#endif

typedef struct PBMSPool_tag{
	int pool_size;	/* The number of connections in the list. */
	PBMSConPtr	con_list[MAX_PBMS_POOL_CONNECTIONS];
	struct PBMSPool_tag *next;
	struct PBMSPool_tag *prev;
} PBMSPool, *PBMSPoolPtr;

static PBMSPoolPtr pools = NULL;
static int le_connection, le_pool;

static int pbms_api_version = 1;

ZEND_DECLARE_MODULE_GLOBALS(pbms)

/*------------------------------------------------------*/
PHP_INI_BEGIN()
	STD_PHP_INI_BOOLEAN("pbms.display_errors",	"1",		PHP_INI_SYSTEM,		OnUpdateLong,		display_errors,		zend_pbms_globals,		pbms_globals)
	STD_PHP_INI_BOOLEAN("pbms.debug_mode",		"0",		PHP_INI_SYSTEM,		OnUpdateLong,		debug_mode,			zend_pbms_globals,		pbms_globals)
	STD_PHP_INI_ENTRY_EX("pbms.max_connections","-1",		PHP_INI_SYSTEM,		OnUpdateLong,		max_connections,	zend_pbms_globals,		pbms_globals,	display_link_numbers)
	STD_PHP_INI_ENTRY_EX("pbms.num_connections","0",		PHP_INI_SYSTEM,		OnUpdateLong,		num_connections,	zend_pbms_globals,		pbms_globals,	display_link_numbers)
	STD_PHP_INI_ENTRY_EX("pbms.pool_size",		"0",		PHP_INI_SYSTEM,		OnUpdateLong,		pool_size,			zend_pbms_globals,		pbms_globals,	display_link_numbers)

	STD_PHP_INI_ENTRY("pbms.default_host",		"localhost",PHP_INI_ALL,		OnUpdateString,		default_host,		zend_pbms_globals,		pbms_globals)
	STD_PHP_INI_ENTRY("pbms.default_port",		"8080",		PHP_INI_ALL,		OnUpdateLong,		default_port,		zend_pbms_globals,		pbms_globals)
	STD_PHP_INI_ENTRY("pbms.default_database",	"",			PHP_INI_ALL,		OnUpdateString,		default_database,	zend_pbms_globals,		pbms_globals)
PHP_INI_END()

/*------------------------------------------------------*/
static void php_pbms_init_globals(zend_pbms_globals *pbms_globals)
{
#ifdef ZTS
	pool_mutex = tsrm_mutex_alloc();
#endif
}

/*------------------------------------------------------*/
static void lock_pools(void)
{
#ifdef ZTS
	tsrm_mutex_lock(pool_mutex);
#endif
}

/*------------------------------------------------------*/
static void unlock_pools(void)
{
#ifdef ZTS
	tsrm_mutex_unlock(pool_mutex);
#endif
}


/*------------------------------------------------------*/
static void _free_pbms_pool(zend_rsrc_list_entry *rsrc TSRMLS_DC)
{
	PBMSPoolPtr pool = (PBMSPoolPtr) rsrc->ptr;
	PBMSConPtr pbcon;
	int i;
	PBMS_LS_FETCH();

	lock_pools();
	for (i = 0; i < pool->pool_size; i++) {
		pbcon = pool->con_list[i];
		if (pbcon) {
			if ( pbcon->pbms && !pbcon->in_use) {
				PBMS_G(num_connections)--;
				pbms_close(pbcon->pbms);
				free(pbcon);
			} else 
				pbcon->my_pool = NULL;
		}
	}
	
	if (pools == pool) 	
		pools = pool->next;
	else
		pool->prev->next = pool->next;
		
	if (pool->next)
		pool->next->prev = pool->prev;
		
	free(pool);
	unlock_pools();
}

/*------------------------------------------------------*/
static void _close_pbms_conn(zend_rsrc_list_entry *rsrc TSRMLS_DC)
{
	PBMSConPtr pbcon = (PBMSConPtr) rsrc->ptr;
	PBMSPoolPtr pool;
	int i;
	PBMS_LS_FETCH();

	lock_pools();
	pool = pbcon->my_pool;
	if (!pool) { /* If the connection isn't in a pool free it.*/
		PBMS_G(num_connections)--;
		pbms_close(pbcon->pbms);
		free(pbcon);
	} else {
		pbcon->in_use = FALSE;
		if (pool->pool_size > PBMS_G(pool_size)) {
			PBMS_G(num_connections)--;
			pbms_close(pbcon->pbms);
			pbcon->pbms = NULL;
			free(pbcon);
			for (i = 0; (i < pool->pool_size) && (pool->con_list[i] != pbcon); i++);
			if (i < pool->pool_size) {
				memmove(pool->con_list +i, pool->con_list +i+1, sizeof(PBMSConPtr) * ( pool->pool_size -i -1));
				pool->pool_size--;
				pool->con_list[pool->pool_size] = NULL;
			}
		}
	}
	

	unlock_pools();
	
}


/*------------------------------------------------------*/
PHP_MINIT_FUNCTION(pbms)
{
	ZEND_INIT_MODULE_GLOBALS(pbms, php_pbms_init_globals, NULL);
	REGISTER_INI_ENTRIES();

	le_connection = zend_register_list_destructors_ex(_close_pbms_conn, NULL, PBMS_CONNECTION_NAME, module_number);
	le_pool = zend_register_list_destructors_ex(NULL, _free_pbms_pool, PBMS_POOL_NAME, module_number);
	
    return (pbms_library_init())? SUCCESS: FAILURE;
}

/*------------------------------------------------------*/
PHP_MSHUTDOWN_FUNCTION(pbms)
{
	pbms_library_end();
	UNREGISTER_INI_ENTRIES();
    return SUCCESS;
}

/*------------------------------------------------------*/
PHP_RINIT_FUNCTION(pbms)
{
	PBMS_LS_FETCH();
	
	PBMS_G(default_connection_id)=-1;
	PBMS_G(num_connections) = 0;
    return SUCCESS;
}

/*------------------------------------------------------*/
PHP_RSHUTDOWN_FUNCTION(pbms)
{
    return SUCCESS;
}

/*------------------------------------------------------*/
PHP_MINFO_FUNCTION(pbms)
{
	char buf[32];
	PBMS_LS_FETCH();

	php_info_print_table_start();
	php_info_print_table_header(2, "PrimeBase Media Streaming (PBMS) Support", "enabled");
	
	php_info_print_table_row(2, "PBMS PHP extension version", PBMS_PHP_VERSION);
	
#ifdef PBMSLIB_API_VERSION
	php_info_print_table_row(2, "PBMS client library version", pbms_get_version());
#endif
	
	snprintf(buf, 32, "%d", PBMS_G(pool_size));
	php_info_print_table_row(2, "Persistent connection pool size", buf);

	snprintf(buf, 32, "%ld", PBMS_G(num_connections));
	php_info_print_table_row(2, "Active Connections", buf);

    php_info_print_table_end();
}

/*------------------------------------------------------*/
static void report_pbms_error(PBMSConPtr pbcon)
{	
	
	pbcon->err_no = pbms_errno(pbcon->pbms);
	strncpy(pbcon->err_msg, pbms_error(pbcon->pbms), MAX_ERROR);
	pbcon->err_msg[MAX_ERROR] = 0;
	
	if (pbcon->display_error)
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "%s", pbcon->err_msg);	
}

/*------------------------------------------------------*/
static void report_php_error(PBMSConPtr pbcon, int err, const char *msg)
{
	if (!pbcon) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "%s", msg);	
		return;
	}
	
	pbcon->err_no = err;
	strncpy(pbcon->err_msg, msg, MAX_ERROR);
	pbcon->err_msg[MAX_ERROR] = 0;
	
	if (pbcon->display_error)
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "%s", pbcon->err_msg);	
}


/*------------------------------------------------------*/
static Bool set_connection(PBMSConPtr pbcon)
{

	if (pbcon->pbms)
		pbms_close(pbcon->pbms);
		
	pbcon->pbms = pbms_connect(pbcon->host, pbcon->port, pbcon->database);
	if (!pbcon->pbms) {
		report_pbms_error(pbcon);
		return FAILED;
	}
	
	return OK;
	
}


/*------------------------------------------------------
 * I do not understand how INTERNAL_FUNCTION_PARAMETERS works and so I do not trust it and therefore do not use it.
 * If I ever found some actual documentation saying how it should and should not be used then I may trust it.
 * The reason I use it is so that I can use the return value setting macros.
 * I hate macros and Zend is macro madness. :(
 */
static void php_pbms_do_connect(INTERNAL_FUNCTION_PARAMETERS, char *host, long port, char *database, int pool_connection)
{
	char *hashed_details;
	int i, hashed_details_length;
  	PBMSConPtr pbcon;
	list_entry *pool_list;
	PBMSPoolPtr pool = NULL;
	
	PBMS_LS_FETCH();

	if ((!host) || !*host) 
		host = PBMS_G(default_host);
	
	if (!port) 
		port = (long)PBMS_G(default_port);
	
	if (!database)
		database = PBMS_G(default_database);
		
	hashed_details_length = strlen("pbmsengine___") + strlen(SAFE_STRING(host)) + strlen(SAFE_STRING(database))+(sizeof(port)*2);
	hashed_details = (char *) emalloc(hashed_details_length+1);
	if (!hashed_details) {
		php_error(E_WARNING, "Out Of Memory!");
		RETURN_FALSE;
	}
	snprintf(hashed_details, hashed_details_length, "pbmsengine_%s_%s_%x", SAFE_STRING(host), SAFE_STRING(database), port);
	hashed_details_length = strlen(hashed_details);
	/* try to find if we already have this connection in our persistent list */
	if ((!PBMS_G(pool_size)) || zend_hash_find(&EG(persistent_list), hashed_details, hashed_details_length+1, (void **) &pool_list)==FAILURE) {  /* we don't */

		if (PBMS_G(max_connections)!=-1 && PBMS_G(num_connections)>=PBMS_G(max_connections)) {
			php_error(E_WARNING, "PBMS:  Too many open connections (%ld)", PBMS_G(num_connections));
			efree(hashed_details);
			RETURN_FALSE;
		}
		
		/* Open the connection */
		pbcon = (PBMSConPtr) malloc(sizeof(PBMSCon));
		if (!pbcon) {
			php_error(E_WARNING, "Out Of Memory!");
			efree(hashed_details);
			RETURN_FALSE;
		}
		memset(pbcon, 0, sizeof(PBMSCon));
		pbcon->display_error =  PBMS_G(display_errors);
		
		if (pool_connection && PBMS_G(pool_size)) {
			pool = (PBMSPoolPtr) malloc(sizeof(PBMSPool));
			if (!pool) {
				php_error(E_WARNING, "Out Of Memory!");
				efree(hashed_details);
				free(pbcon);
				RETURN_FALSE;
			}
			memset(pool, 0, sizeof(PBMSPool));
		}
		
		if (host)
			strncpy(pbcon->host, host, PBMS_MAX_HOST_SIZE);
		if (database) 
			strcpy(pbcon->database, database);
	
		pbcon->port = port;

		if (!set_connection(pbcon)) {
			free(pbcon);
			if (pool) free(pool);
			efree(hashed_details);
			RETURN_FALSE;
		}
		PBMS_G(num_connections)++;
		
		
		pbcon->in_use = TRUE;
		pbcon->my_pool = pool;

		if (pool) {
			list_entry new_pool;
			
			pool->con_list[0] = pbcon;
			pool->pool_size = 1;
			
			
			/* hash it up */
			new_pool.type = le_pool;
			new_pool.ptr = pool;
			if (zend_hash_update(&EG(persistent_list), hashed_details, hashed_details_length+1, (void *) &new_pool, sizeof(list_entry), NULL)==FAILURE) {
				PBMS_G(num_connections)--;
				pbms_close(pbcon->pbms);
				free(pbcon);
				free(pool);
				efree(hashed_details);
				RETURN_FALSE;
			}
			
			lock_pools();
			pool->next = pools;
			pools = pool;
			unlock_pools();
			
		}
	} else {  /* we do */
		if (pool_list->type != le_pool) {
			RETURN_FALSE;
		}
		
		pool = (PBMSPoolPtr) pool_list->ptr;
		pbcon = NULL;
		lock_pools();
		for (i = 0; i < pool->pool_size && pool->con_list[i]->in_use; i++);
		if ( i < pool->pool_size) {
			pbcon = pool->con_list[i];
			pbcon->in_use = TRUE;
		}
		unlock_pools();
		
		if (!pbcon) {
			if (PBMS_G(max_connections)!=-1 && PBMS_G(num_connections)>=PBMS_G(max_connections)) {
				php_error(E_WARNING, "PBMS:  Too many open connections (%ld)", PBMS_G(num_connections));
				efree(hashed_details);
				RETURN_FALSE;
			}
			
			if (pool->pool_size == MAX_PBMS_POOL_CONNECTIONS) {
				php_error(E_WARNING, "Maximum connections per server exceeded! (As defined by the compiler flag MAX_PBMS_POOL_CONNECTIONS.)");
				efree(hashed_details);
				RETURN_FALSE;
			}
			
			pbcon = (PBMSConPtr) malloc(sizeof(PBMSCon));
			if (!pbcon) {
				php_error(E_WARNING, "Out Of Memory!");
				efree(hashed_details);
				RETURN_FALSE;
			}
			memset(pbcon, 0, sizeof(PBMSCon));
			pbcon->display_error =  PBMS_G(display_errors);
			
			pbcon->in_use = TRUE;
			if (host)
				strncpy(pbcon->host, host, PBMS_MAX_HOST_SIZE);
				
			if (database) 
				strcpy(pbcon->database, database);
		
			pbcon->port = port;
			
			lock_pools();
			if (pool_connection && (pool->pool_size < PBMS_G(pool_size))) {			
				pbcon->my_pool = pool;
				pool->con_list[pool->pool_size] = pbcon;
				pool->pool_size++;
			} else {
				pbcon->my_pool = NULL;
			}
			unlock_pools();
		}
		
		/* ensure that the link did not die */
		if (!pbcon->pbms){ /* The connection died. */

			if (!set_connection(pbcon)) {
				pbcon->in_use = FALSE;
				if (!pbcon->my_pool)
					free(pbcon);
				efree(hashed_details);
				RETURN_FALSE;
			}
			
			PBMS_G(num_connections)++;
		}
			
	}
	
	ZEND_REGISTER_RESOURCE(return_value, pbcon, le_connection);

	efree(hashed_details);
	
	if (PBMS_G(default_connection_id)!= -1) {
		zend_list_delete(PBMS_G(default_connection_id)); /* Removes a reference to the link */
	}
	PBMS_G(default_connection_id) = return_value->value.lval;
	zend_list_addref(return_value->value.lval); /* Add a reference so the link isn't deleted prematurely.*/
}

/*------------------------------------------------------*/
static PBMSConPtr php_pbms_get_default_connection(INTERNAL_FUNCTION_PARAMETERS, Bool create)
{
	zval **zpbcon=NULL;
	PBMSConPtr pbcon = NULL;
	int pbms_res_type;
	PBMS_LS_FETCH();

	
	if (PBMS_G(default_connection_id)== -1) { /* no link opened yet, implicitly open one */
		if (!create)
			return NULL;
		php_pbms_do_connect(INTERNAL_FUNCTION_PARAM_PASSTHRU, NULL, 0, NULL, 0);
	}
	
	ZEND_FETCH_RESOURCE(pbcon, PBMSConPtr, zpbcon, PBMS_G(default_connection_id), PBMS_CONNECTION_NAME, le_connection);

	return pbcon;
}

/*------------------------------------------------------*/
/* {{{ proto long pbms_errno([resource connection_identifier])
	For the connection specified by $pbms, returns the error code for the most 
	recently invoked API function that can succeed or fail.  */
PHP_FUNCTION(pbms_errno)
{
    PBMSConPtr pbcon;
    zval *zpbcon = NULL;
 	int pbms_res_type;
	int err = PHP_PBMS_CONID_ERR;

    if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "|r", &zpbcon)){
        RETURN_FALSE;
	}

    if (zpbcon) {
        pbcon = (PBMSConPtr) zend_list_find(Z_RESVAL_P(zpbcon), &pbms_res_type);
		if (pbms_res_type != le_connection) 
			pbcon = NULL;
	} else 
		pbcon = php_pbms_get_default_connection(INTERNAL_FUNCTION_PARAM_PASSTHRU, FALSE);
	
	if (pbcon) {
		err = pbcon->err_no;
	}
	
   RETURN_LONG(err);
}
/* }}} */

/*------------------------------------------------------*/
/* {{{ proto string pbms_error([resource connection_identifier])
	For the connection specified by connection_identifier, returns a null-terminated 
	string containing the error message for the most recently invoked API function 
	that can succeed or fail. */
PHP_FUNCTION(pbms_error)
{
    PBMSConPtr pbcon;
    zval *zpbcon = NULL;
 	int pbms_res_type;
	char *err_msg = "Invalid PBMS Connection ID";

    if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "|r", &zpbcon)){
        RETURN_FALSE;
	}

    if (zpbcon) {
        pbcon = (PBMSConPtr) zend_list_find(Z_RESVAL_P(zpbcon), &pbms_res_type);
		if (pbms_res_type != le_connection) 
			pbcon = NULL;
	} else 
		pbcon = php_pbms_get_default_connection(INTERNAL_FUNCTION_PARAM_PASSTHRU, FALSE);
	
	if (pbcon) {
		err_msg = pbcon->err_msg;
	}
	

    RETURN_STRING(err_msg, 1);
}
/* }}} */

//------------------------------------------------------
static void reset_pool_sizes(int new_size)
{
	PBMSPoolPtr pool = NULL;
	PBMSConPtr pbcon;
	int i;
	
	PBMS_LS_FETCH();

	lock_pools();
	
	pool = pools;
	
	while (pool) {
		if (pool->pool_size > PBMS_G(pool_size)) {;
			for (i = 0; (pool->pool_size > new_size)  && (i < pool->pool_size) ;){
				pbcon = pool->con_list[i];
				if (! pbcon->in_use) {
					memmove(pool->con_list +i, pool->con_list +i+1, sizeof(PBMSConPtr) * ( pool->pool_size -i -1));
					pool->pool_size--;
					pool->con_list[pool->pool_size] = NULL;
					
					PBMS_G(num_connections)--;
					pbms_close(pbcon->pbms);
					free(pbcon);
				} else 
					i++;
			}
		}
		pool = pool->next;
	}
	
	unlock_pools();
}

//------------------------------------------------------------------------------------
/* {{{ proto int pbms_connection_pool_size(int pool_size)
   Open a connection to a PrimeBase Server */
PHP_FUNCTION(pbms_connection_pool_size)
{
	int old_size, new_size;
	PBMS_LS_FETCH();
	
    if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "l", &new_size)){
        RETURN_FALSE;
	}

	old_size = PBMS_G(pool_size);
	PBMS_G(pool_size) = new_size;
	
	if (new_size < old_size) // Close any unused connections.
		reset_pool_sizes(new_size);
		
	RETURN_TRUE;
}
/* }}} */

/*------------------------------------------------------*/
/* {{{ proto resource pbms_pconnect([string host[, long port[, string database]]]])
	Establishes a persistent connection to a PBMS BLOB streaming server running on $host.*/
PHP_FUNCTION(pbms_pconnect)
{
	char *host = NULL, *database = NULL;
	int host_len, database_len;
	long port = 0;

    if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "|sls", &host, &host_len, &port, &database, &database_len)){
        RETURN_FALSE;
	}

	php_pbms_do_connect(INTERNAL_FUNCTION_PARAM_PASSTHRU PBMS_LS_DC, host, port, database, 1);
}
/* }}} */

/*------------------------------------------------------*/
/* {{{ proto resource pbms_connect([string host[, long port[, string database]]]])
	Establishes a non persistent connection to a PBMS BLOB streaming server running on $host.*/
PHP_FUNCTION(pbms_connect)
{
	char *host = NULL, *database = NULL;
	int host_len, database_len;
	long port = 0;

    if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS()  TSRMLS_CC, "|sls", &host, &host_len, &port, &database, &database_len)) {
        RETURN_FALSE;
	}
	
	php_pbms_do_connect(INTERNAL_FUNCTION_PARAM_PASSTHRU PBMS_LS_DC, host, port, database, 0);
}
/* }}} */

/*------------------------------------------------------*/
/* {{{ proto void pbms_print_con_info()
	An undocumented function used for testing to dump connection info. */
PHP_FUNCTION(pbms_print_con_info)
{
	int i, num_pools = 0, pool_active = 0, total_pool_cons = 0;
	PBMSPoolPtr pool = NULL;
	PBMS_LS_FETCH();

	php_printf("Connection total: %d, Pool size: %d\n", (int)PBMS_G(num_connections), (int)PBMS_G(pool_size));
	lock_pools();
	
	pool = pools;
	while (pool) {
		num_pools++;
		pool_active = 0;
		for (i = 0; i < pool->pool_size ; i++){
			if (pool->con_list[i]->in_use) 
				pool_active++;
		}
		
		php_printf("Pool: %d, Connected: %d, Active: %d \n", num_pools, pool->pool_size, pool_active);
		total_pool_cons += pool->pool_size;
		
		pool = pool->next;
	}
	
	unlock_pools();

	php_printf("Non Pool Connections: %d\n", (int)(PBMS_G(num_connections) - total_pool_cons));

}
/* }}} */

/*------------------------------------------------------*/
/* {{{ proto void pbms_close([resource connection_identifier])
	Closes the connection to the PBMS BLOB streaming server associated with the 
	PBMS connection resource and frees all of it's resources. */
PHP_FUNCTION(pbms_close)
{
    PBMSConPtr pbcon;
    zval *zpbcon = NULL;
 	int pbms_res_type;
	PBMS_LS_FETCH();

    if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "|r", &zpbcon)){
        RETURN_FALSE;
	}

    if (zpbcon) {
        pbcon = (PBMSConPtr) zend_list_find(Z_RESVAL_P(zpbcon), &pbms_res_type);
		if (pbms_res_type != le_connection) {
			php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid Resource Provided");
			RETURN_FALSE;
		}
		
	} else 
		pbcon = php_pbms_get_default_connection(INTERNAL_FUNCTION_PARAM_PASSTHRU, FALSE);

	if (pbcon) {
		/* 
		 * In case this is a pool connection clear the metadata so it is ready
		 * for the next session.
		 */
		pbms_clear_metadata(pbcon->pbms, NULL);

		CLEAR_ERRORS(pbcon);
		if (zpbcon) {
			zend_list_delete(Z_RESVAL_P(zpbcon));
			if (Z_RESVAL_P(zpbcon) == PBMS_G(default_connection_id))
				PBMS_G(default_connection_id) = -1;
		} else {
			zend_list_delete(PBMS_G(default_connection_id));
			PBMS_G(default_connection_id) = -1;
		}
	}
	RETURN_TRUE
}
/* }}} */

/*------------------------------------------------------*/
/* {{{ proto void pbms_add_metadata(string name, string value, [resource connection_identifier])
	Adds the metadata to the connection to be associated with the next BLOB uploaded to the 
	PBMS BLOB streaming server.

	The metadata remains associated with the connection until it is cleared by calling 
	pbms_clear_metadata().

	The metadata $name must exist in the PBMS system table 'pbms_metadata_header' before the 
	PBMS BLOB streaming server will store it with the BLOB data. Each database has it's own 
	'pbms_metadata_header' table which is created automatically the first time it is referenced. */
PHP_FUNCTION(pbms_add_metadata)
{
    PBMSConPtr pbcon;
    zval *zpbcon = NULL;
 	int pbms_res_type;
    int name_len;
    int value_len;
    const char *name = NULL;
    const char *value = NULL;

    if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ss|r", &name, &name_len, &value, &value_len, &zpbcon)){
        RETURN_FALSE;
	}

    if (zpbcon) {
        pbcon = (PBMSConPtr) zend_list_find(Z_RESVAL_P(zpbcon), &pbms_res_type);
		if (pbms_res_type != le_connection) 
			pbcon = NULL;
	} else 
		pbcon = php_pbms_get_default_connection(INTERNAL_FUNCTION_PARAM_PASSTHRU, TRUE);
	
	if (!pbcon) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid Resource Provided");
		RETURN_FALSE;
	}
	
	CLEAR_ERRORS(pbcon);
	
	if (!pbms_add_metadata(pbcon->pbms, name, value)) {
		report_pbms_error(pbcon);
		RETURN_FALSE;
	}
	

	RETURN_TRUE;
}
/* }}} */

/*------------------------------------------------------*/
/* {{{ proto void pbms_clear_metadata([string name [,resource connection_identifier]])
	Clears the metadata name and value for $name  from the connection. If 'name' 
	is not supplied then all metadata is removed from the connection.  */
PHP_FUNCTION(pbms_clear_metadata)
{
    PBMSConPtr pbcon;
    zval *zpbcon = NULL;
	int name_len;
	const char *name = NULL;
	int pbms_res_type;

    if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "|sr", &name, &name_len, &zpbcon)){
        RETURN_FALSE;
	}

    if (zpbcon) {
        pbcon = (PBMSConPtr) zend_list_find(Z_RESVAL_P(zpbcon), &pbms_res_type);
		if (pbms_res_type != le_connection) 
			pbcon = NULL;
	} else 
		pbcon = php_pbms_get_default_connection(INTERNAL_FUNCTION_PARAM_PASSTHRU, TRUE);
	
	if (!pbcon) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid Resource Provided");
		RETURN_FALSE;
	}
	
	CLEAR_ERRORS(pbcon);
		
	pbms_clear_metadata(pbcon->pbms, name);
	
    RETURN_TRUE;
}
/* }}} */

/*------------------------------------------------------*/
/* {{{ proto long pbms_reset_metadata([resource connection_identifier])
	Resets the position of the metadata cursor to the start of the 
	metadata for the last BLOB or BLOB info retrieved, returning 
	the number of metadata fields associated with the BLOB. */
PHP_FUNCTION(pbms_reset_metadata)
{
    PBMSConPtr pbcon;
    zval *zpbcon = NULL;
 	int pbms_res_type;

    if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "|r", &zpbcon)){
       RETURN_FALSE;
	}

    if (zpbcon) {
        pbcon = (PBMSConPtr) zend_list_find(Z_RESVAL_P(zpbcon), &pbms_res_type);
		if (pbms_res_type != le_connection) 
			pbcon = NULL;
	} else 
		pbcon = php_pbms_get_default_connection(INTERNAL_FUNCTION_PARAM_PASSTHRU, TRUE);
	
	if (!pbcon) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid Resource Provided");
		RETURN_FALSE;
	}
	
	CLEAR_ERRORS(pbcon);

    RETURN_LONG(pbms_reset_metadata(pbcon->pbms));
}
/* }}} */

/*------------------------------------------------------*/
/* {{{ proto (string,string) pbms_next_metadata([resource connection_identifier])
	Fetches the next name/value pair from the metadata cursor while advancing the 
	cursor to the next row. The return value is an array of two strings, the first 
	is the metadata name and the second is the value associated with the name.   */
PHP_FUNCTION(pbms_next_metadata)
{
	PBMSConPtr pbcon;
    zval *zpbcon = NULL;
 	int pbms_res_type;
	char name[MS_META_NAME_SIZE];
	char *value = NULL;
	size_t v_size, size;
	Bool has_metadata;

    if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "|r", &zpbcon)){
        RETURN_FALSE;
	}

    if (zpbcon) {
        pbcon = (PBMSConPtr) zend_list_find(Z_RESVAL_P(zpbcon), &pbms_res_type);
		if (pbms_res_type != le_connection) 
			pbcon = NULL;
	} else 
		pbcon = php_pbms_get_default_connection(INTERNAL_FUNCTION_PARAM_PASSTHRU, TRUE);
	
	if (!pbcon) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid Resource Provided");
		RETURN_FALSE;
	}

	CLEAR_ERRORS(pbcon);

#ifdef DEBUG
	v_size = 5;
#else
	v_size = 128;
#endif

	do {
		if (value)
			value = (char*) erealloc(value, v_size);
		else
			value = (char*) emalloc(v_size);
		if (!value) {
			report_php_error(pbcon, PHP_PBMS_MEMORY_ERR, "Out of memory");
			RETURN_FALSE;
		}
		
		size = v_size;
		has_metadata = pbms_next_metadata(pbcon->pbms, name, value, &v_size);
	} while (size < v_size);

	if (!has_metadata) {
		if (pbms_errno(pbcon->pbms)) {
			report_pbms_error(pbcon);
			RETURN_FALSE;
		}
		*name = 0;
		*value = 0;
	}
	
	/* Setup the return value as an array.*/
	array_init(return_value);
	add_next_index_string(return_value, name, 1);
	add_next_index_string(return_value, value, 0);
	
}
/* }}} */

/*------------------------------------------------------*/
/* {{{ proto string pbms_get_metadata_value(string name [, resource connection_identifier])
	Gets the value for metadata field 'name'  from the connection. If there is more 
	than one field with the same name then which value is returned by this function 
	is undefined.  */
PHP_FUNCTION(pbms_get_metadata_value)
{
	PBMSConPtr pbcon;
    zval *zpbcon = NULL;
 	int pbms_res_type;
	char *name;
    int name_len;
	char *value = NULL;
	size_t v_size, size;
	Bool has_metadata;

    if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|r", &name, &name_len, &zpbcon)){
       RETURN_FALSE;
	}

    if (zpbcon) {
        pbcon = (PBMSConPtr) zend_list_find(Z_RESVAL_P(zpbcon), &pbms_res_type);
		if (pbms_res_type != le_connection) 
			pbcon = NULL;
	} else 
		pbcon = php_pbms_get_default_connection(INTERNAL_FUNCTION_PARAM_PASSTHRU, TRUE);
	
	if (!pbcon) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid Resource Provided");
		RETURN_FALSE;
	}

	CLEAR_ERRORS(pbcon);

#ifdef DEBUG
	v_size = 5;
#else
	v_size = 128;
#endif

	do {
		if (value)
			value = (char*) erealloc(value, v_size);
		else
			value = (char*) emalloc(v_size);
		if (!value) {
			report_php_error(pbcon, PHP_PBMS_MEMORY_ERR, "Out of memory");
			RETURN_FALSE;
		}
		
		size = v_size;
		has_metadata = pbms_get_metadata_value(pbcon->pbms, name, value, &v_size);
	} while (size < v_size);

	if (!has_metadata) {
		if (pbms_errno(pbcon->pbms)) {
			report_pbms_error(pbcon);
			RETURN_FALSE;
		}
		*value = 0;
	}

    RETURN_STRING(value, 0);
}
/* }}} */

/*------------------------------------------------------*/
/* {{{ proto string pbms_get_md5_digest([ resource connection_identifier])
	Gets the MD5 digest of the BLOB just downloaded by pbms_get_data(), 
	pbms_write_stream(), or pbms_get_info(). */
PHP_FUNCTION(pbms_get_md5_digest)
{
	PBMSConPtr pbcon;
    zval *zpbcon = NULL;
 	int pbms_res_type;
	char *md5_digest;
 
    if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "|r", &zpbcon)){
        RETURN_FALSE;
	}

    if (zpbcon) {
        pbcon = (PBMSConPtr) zend_list_find(Z_RESVAL_P(zpbcon), &pbms_res_type);
		if (pbms_res_type != le_connection) 
			pbcon = NULL;
	} else 
		pbcon = php_pbms_get_default_connection(INTERNAL_FUNCTION_PARAM_PASSTHRU, TRUE);
	
	if (!pbcon) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid Resource Provided");
		RETURN_FALSE;
	}

	CLEAR_ERRORS(pbcon);

	md5_digest = (char *) emalloc(16);
	if (!pbms_get_md5_digest(pbcon->pbms, md5_digest)) {
		efree(md5_digest);
		report_pbms_error(pbcon);
		RETURN_FALSE;
	}
	
	return_value->value.str.len = 16;
   	return_value->value.str.val = md5_digest;
   	return_value->type = IS_STRING;	
}
/* }}} */

/*------------------------------------------------------*/
/* {{{ proto void pbms_get_info(string blob_ref [, resource connection_identifier])
	Get metadata and headers from the PBMS BLOB streaming daemon for the 
	specified BLOB reference.*/
PHP_FUNCTION(pbms_get_info)
{
	PBMSConPtr pbcon;
    zval *zpbcon = NULL;
 	int pbms_res_type;
	char *blob_ref;
	int blob_ref_len;
  
    if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|r", &blob_ref, &blob_ref_len, &zpbcon)){
        RETURN_FALSE;
	}

    if (zpbcon) {
        pbcon = (PBMSConPtr) zend_list_find(Z_RESVAL_P(zpbcon), &pbms_res_type);
		if (pbms_res_type != le_connection) 
			pbcon = NULL;
	} else 
		pbcon = php_pbms_get_default_connection(INTERNAL_FUNCTION_PARAM_PASSTHRU, TRUE);
	
	if (!pbcon) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid Resource Provided");
		RETURN_FALSE;
	}

	CLEAR_ERRORS(pbcon);

	if (!pbms_get_info(pbcon->pbms, blob_ref)) {
		report_pbms_error(pbcon);
		RETURN_FALSE;
	}
	
    RETURN_TRUE;
}
/* }}} */

typedef struct {
	php_stream *stream;
	char *error_msg;
} CBDataRec, *CBDataPtr;

/*------------------------------------------------------*/
static size_t my_stream_write(void *caller_data, const char *buffer, size_t size, Bool reset)
{
	CBDataPtr data = (CBDataPtr) caller_data;
	size_t cnt;
	
	if (reset) {
		if (php_stream_seek(data->stream, 0, SEEK_SET)) {
			data->error_msg = "PBMS: my_stream_write(): php_stream_seek() error";
			return 0;
		}
	}
	cnt = php_stream_write(data->stream, buffer, size);
	if (cnt < size) 
		data->error_msg = "PBMS: my_stream_write(): php_stream_write() error";
		
	return cnt;
}

/*------------------------------------------------------*/
/* {{{ proto void pbms_write_stream(php_stream *output_stream, string blob_ref [, resource connection_identifier])
	Get data from the PBMS BLOB streaming server for the specified BLOB reference.

	The BLOB data is written to the provided stream as it is read in from the PBMS BLOB streaming server. 
	The stream can be any type of stream that supports the 'write' operation.*/
PHP_FUNCTION(pbms_write_stream)
{
	PBMSConPtr pbcon;
    zval *zpbcon = NULL;
	char buffer[TRANSFER_BUFFER_SIZE];
	zval *zstream;
 	int pbms_res_type;
	char *blob_ref;
	int blob_ref_len;
 	CBDataRec cbData = {0};
 
    if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "zs|r", &zstream, &blob_ref, &blob_ref_len, &zpbcon)){
        RETURN_FALSE;
	}

    if (zpbcon) {
        pbcon = (PBMSConPtr) zend_list_find(Z_RESVAL_P(zpbcon), &pbms_res_type);
		if (pbms_res_type != le_connection) 
			pbcon = NULL;
	} else 
		pbcon = php_pbms_get_default_connection(INTERNAL_FUNCTION_PARAM_PASSTHRU, TRUE);
	
	if (!pbcon) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid Resource Provided");
		RETURN_FALSE;
	}


	CLEAR_ERRORS(pbcon);

    php_stream_from_zval(cbData.stream, &zstream);
	cbData.error_msg = NULL;
	
	if ((!pbms_get_data_cb(pbcon->pbms, blob_ref, my_stream_write, &cbData)) || cbData.error_msg) {
		if (cbData.error_msg) 
			report_php_error(pbcon, PHP_PBMS_CALLBACK_ERR, cbData.error_msg);
		else
			report_pbms_error(pbcon);
		RETURN_FALSE;
	}

    RETURN_TRUE;
}
/* }}} */

/*------------------------------------------------------*/
/* {{{ proto void pbms_write_stream_range(php_stream *output_stream, string blob_ref, long first_byte, long last_byte [, resource connection_identifier])
	Get a portion of a BLOB from the PBMS BLOB streaming server for the specified BLOB reference.

	The BLOB data is written to the provided stream as it is read in from the PBMS BLOB streaming server. 
	The stream can be any type of stream that supports the 'write' operation.*/
PHP_FUNCTION(pbms_write_stream_range)
{
	PBMSConPtr pbcon;
    zval *zpbcon = NULL;
	char buffer[TRANSFER_BUFFER_SIZE];
	zval *zstream;
 	int pbms_res_type;
	char *blob_ref;
	int blob_ref_len;
	long first_byte, last_byte;
 	CBDataRec cbData = {0};
 
    if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "zsll|r", &zstream, &blob_ref, &blob_ref_len, &first_byte, &last_byte, &zpbcon)){
        RETURN_FALSE;
	}

    if (zpbcon) {
        pbcon = (PBMSConPtr) zend_list_find(Z_RESVAL_P(zpbcon), &pbms_res_type);
		if (pbms_res_type != le_connection) 
			pbcon = NULL;
	} else 
		pbcon = php_pbms_get_default_connection(INTERNAL_FUNCTION_PARAM_PASSTHRU, TRUE);
	
	if (!pbcon) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid Resource Provided");
		RETURN_FALSE;
	}

	if (first_byte > last_byte) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid Range Provided");
		RETURN_FALSE;
	}

	CLEAR_ERRORS(pbcon);

    php_stream_from_zval(cbData.stream, &zstream);
	cbData.error_msg = NULL;
	
	if ((!pbms_get_data_range_cb(pbcon->pbms, blob_ref, first_byte, last_byte, my_stream_write, &cbData)) || cbData.error_msg) {
		if (cbData.error_msg) 
			report_php_error(pbcon, PHP_PBMS_CALLBACK_ERR, cbData.error_msg);
		else
			report_pbms_error(pbcon);
		RETURN_FALSE;
	}

    RETURN_TRUE;
}
/* }}} */

/*------------------------------------------------------*/
/* {{{ proto string pbms_get_data(string blob_ref [, resource connection_identifier])
	Get data from the PBMS BLOB streaming server for the specified BLOB reference. */
PHP_FUNCTION(pbms_get_data)
{
	PBMSConPtr pbcon;
    zval *zpbcon = NULL;
 	int pbms_res_type;
	char *blob_ref;
	unsigned char * buffer;
	int blob_ref_len;
	size_t size;
  
    if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|r", &blob_ref, &blob_ref_len, &zpbcon)){
        RETURN_FALSE;
	}

    if (zpbcon) {
        pbcon = (PBMSConPtr) zend_list_find(Z_RESVAL_P(zpbcon), &pbms_res_type);
		if (pbms_res_type != le_connection) 
			pbcon = NULL;
	} else 
		pbcon = php_pbms_get_default_connection(INTERNAL_FUNCTION_PARAM_PASSTHRU, TRUE);
	
	if (!pbcon) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid Resource Provided");
		RETURN_FALSE;
	}

	CLEAR_ERRORS(pbcon);

	if (!pbms_get_blob_size(pbcon->pbms, blob_ref, &size)){
		report_pbms_error(pbcon);
		RETURN_FALSE;
	}

	buffer = (unsigned char *) emalloc(size);
	if (!buffer) {
		report_php_error(pbcon, PHP_PBMS_MEMORY_ERR, "Out of memory");
		RETURN_FALSE;
	}
	
	if (!pbms_get_data(pbcon->pbms, blob_ref, buffer, size)) {
		report_pbms_error(pbcon);
		efree(buffer);
		RETURN_FALSE;
	}
	
	return_value->value.str.len = size;
   	return_value->value.str.val = (char*) buffer;
   	return_value->type = IS_STRING;
}
/* }}} */

/*------------------------------------------------------*/
/* {{{ proto string pbms_get_data_range(string blob_ref, long first_byte, long last_byte [, resource connection_identifier])
	Get a portion of a BLOB from the PBMS BLOB streaming server for the specified BLOB reference.*/
PHP_FUNCTION(pbms_get_data_range)
{
	PBMSConPtr pbcon;
    zval *zpbcon = NULL;
 	int pbms_res_type;
	char *blob_ref;
	unsigned char * buffer;
	int blob_ref_len;
	size_t size, data_size =0;
  	long first_byte, last_byte;

    if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "sll|r", &blob_ref, &blob_ref_len, &first_byte, &last_byte, &zpbcon)){
        RETURN_FALSE;
	}

    if (zpbcon) {
        pbcon = (PBMSConPtr) zend_list_find(Z_RESVAL_P(zpbcon), &pbms_res_type);
		if (pbms_res_type != le_connection) 
			pbcon = NULL;
	} else 
		pbcon = php_pbms_get_default_connection(INTERNAL_FUNCTION_PARAM_PASSTHRU, TRUE);
	
	if (!pbcon) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid Resource Provided");
		RETURN_FALSE;
	}

	if (first_byte > last_byte) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid Range Provided");
		RETURN_FALSE;
	}
	
	CLEAR_ERRORS(pbcon);

	if (!pbms_get_blob_size(pbcon->pbms, blob_ref, &size)){
		report_pbms_error(pbcon);
		RETURN_FALSE;
	}

	if (size > (last_byte - first_byte))
		size = last_byte - first_byte +1;
		
	buffer = (unsigned char *) emalloc(size);
	if (!buffer) {
		report_php_error(pbcon, PHP_PBMS_MEMORY_ERR, "Out of memory");
		RETURN_FALSE;
	}
	
	if (!pbms_get_data_range(pbcon->pbms, blob_ref, first_byte, last_byte, buffer, size, &data_size)) {
		report_pbms_error(pbcon);
		efree(buffer);
		RETURN_FALSE;
	}
	
	return_value->value.str.len = data_size;
   	return_value->value.str.val = (char*) buffer;
   	return_value->type = IS_STRING;
}
/* }}} */

/*------------------------------------------------------*/
static size_t my_stream_read(void *caller_data, char *buffer, size_t size, Bool reset)
{
	CBDataPtr data = (CBDataPtr) caller_data;
	size_t cnt;
	
	if (reset) {
		if (php_stream_seek(data->stream, 0, SEEK_SET)) {
			data->error_msg = "PBMS: my_stream_read(): php_stream_seek() error";
			return 0;
		}
	}

	cnt = php_stream_read(data->stream, buffer, size);
	if ((cnt < size) && (php_stream_eof(data->stream) != 1)) {
		data->error_msg = "PBMS: my_stream_read(): php_stream_read() error";
	}
		
	return cnt;
}

/*------------------------------------------------------*/
/* {{{ proto string pbms_read_stream(php_stream *input_stream, int size[, string checksum[, resource connection_identifier]])
   Sends data to the PBMS BLOB streaming server returning a BLOB reference. The BLOB reference can then 
   be inserted into a longblob column of a PBMS enabled table inplace of the actual BLOB data.

   The caller provides an open stream from which the BLOB data to be sent is read. The stream can be any 
   type of stream that supports the 'read' operation.  */
PHP_FUNCTION(pbms_read_stream)
{
	PBMSConPtr pbcon;
	CBDataRec cbData = {0};
	php_stream *stream;
    zval *zpbcon = NULL;
	char buffer[TRANSFER_BUFFER_SIZE];
	char blob_ref[MS_BLOB_URL_SIZE];
	zval *zstream;
	zval *zchecksum = NULL;
	char *checksum = NULL;
	
	int pbms_res_type;
	int size;
    
	if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rl|z!r", &zstream, &size, &zchecksum, &zpbcon)){
        RETURN_FALSE;
	}

    if (zpbcon) {
        pbcon = (PBMSConPtr) zend_list_find(Z_RESVAL_P(zpbcon), &pbms_res_type);
		if (pbms_res_type != le_connection) 
			pbcon = NULL;
	} else 
		pbcon = php_pbms_get_default_connection(INTERNAL_FUNCTION_PARAM_PASSTHRU, TRUE);
	
	if (!pbcon) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid Resource Provided");
		RETURN_FALSE;
	}

	CLEAR_ERRORS(pbcon);

    php_stream_from_zval(cbData.stream, &zstream);
	cbData.error_msg = NULL;
	
 	if (zchecksum)
		checksum = Z_STRVAL_P(zchecksum);

	if ((!pbms_put_data_cb(pbcon->pbms, checksum, blob_ref, size, my_stream_read, &cbData)) || cbData.error_msg) {
		if (cbData.error_msg) 
			report_php_error(pbcon, PHP_PBMS_CALLBACK_ERR, cbData.error_msg);
		else
			report_pbms_error(pbcon);
		RETURN_FALSE;
	}

    RETURN_STRING(blob_ref, 1);
}
/* }}} */

/*------------------------------------------------------*/
/* {{{ proto string pbms_put_data( string data [, string checksum [, resource connection_identifier]])
   Sends data to the PBMS BLOB streaming server returning a BLOB reference. The BLOB reference can then be inserted 
   into a longblob column of a PBMS enabled table inplace of the actual BLOB data. */
PHP_FUNCTION(pbms_put_data)
{
	PBMSConPtr pbcon;
    zval *zpbcon = NULL;
	char blob_ref[MS_BLOB_URL_SIZE];
	zval *zchecksum = NULL;
	char *checksum = NULL;
	unsigned char *data = NULL;	
	int pbms_res_type;
	int size;
       
	if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|z!r", &data, &size, &zchecksum, &zpbcon)){
        RETURN_FALSE;
	}

    if (zpbcon) {
        pbcon = (PBMSConPtr) zend_list_find(Z_RESVAL_P(zpbcon), &pbms_res_type);
		if (pbms_res_type != le_connection) 
			pbcon = NULL;
	} else 
		pbcon = php_pbms_get_default_connection(INTERNAL_FUNCTION_PARAM_PASSTHRU, TRUE);
	
	if (!pbcon) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid Resource Provided");
		RETURN_FALSE;
	}

	CLEAR_ERRORS(pbcon);

 	if (zchecksum)
		checksum = Z_STRVAL_P(zchecksum);

	if (!pbms_put_data(pbcon->pbms, checksum, blob_ref, size, data)) {
		report_pbms_error(pbcon);
		RETURN_FALSE;
	}

    RETURN_STRING(blob_ref, 1);
}
/* }}} */

/*------------------------------------------------------*/
/* {{{ proto bool pbms_is_blob_reference(string blob_ref[, resource connection_identifier])
   Checks the format of the BLOB reference and returns true if it looks like a valid reference. 
   This doesn't guarantee that the reference references a valid BLOB on the PBMS BLOB streaming 
   server only that the format of the BLOB reference is correct. */
PHP_FUNCTION(pbms_is_blob_reference)
{
	PBMSConPtr pbcon;
    zval *zpbcon = NULL;
 	int pbms_res_type;
	char *blob_ref;
	int blob_ref_len;
	Bool is_ref = FALSE;
  
    if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|r", &blob_ref, &blob_ref_len, &zpbcon)){
        RETURN_FALSE;
	}

    if (zpbcon) {
        pbcon = (PBMSConPtr) zend_list_find(Z_RESVAL_P(zpbcon), &pbms_res_type);
		if (pbms_res_type != le_connection) 
			pbcon = NULL;
	} else 
		pbcon = php_pbms_get_default_connection(INTERNAL_FUNCTION_PARAM_PASSTHRU, TRUE);
	
	if (!pbcon) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid Resource Provided");
		RETURN_FALSE;
	}

	CLEAR_ERRORS(pbcon);

	if (!pbms_is_blob_reference(pbcon->pbms, blob_ref)) {
		if (pbms_errno(pbcon->pbms)) {
			report_pbms_error(pbcon);
			RETURN_FALSE;
		}
	} else
		is_ref = TRUE;

    RETVAL_BOOL(is_ref);
}
/* }}} */

/*------------------------------------------------------*/
/* {{{ proto long pbms_get_blob_size(string blob_ref[, resource connection_identifier])
   Get the size of the BLOB for the specified BLOB reference. */
PHP_FUNCTION(pbms_get_blob_size)
{
	PBMSConPtr pbcon;
    zval *zpbcon = NULL;
 	int pbms_res_type;
	char *blob_ref;
	int blob_ref_len;
	size_t size;
  
    if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|r", &blob_ref, &blob_ref_len, &zpbcon)){
        RETURN_FALSE;
	}

    if (zpbcon) {
        pbcon = (PBMSConPtr) zend_list_find(Z_RESVAL_P(zpbcon), &pbms_res_type);
		if (pbms_res_type != le_connection) 
			pbcon = NULL;
	} else 
		pbcon = php_pbms_get_default_connection(INTERNAL_FUNCTION_PARAM_PASSTHRU, TRUE);
	
	if (!pbcon) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid Resource Provided");
		RETURN_FALSE;
	}

	CLEAR_ERRORS(pbcon);

	if (!pbms_get_blob_size(pbcon->pbms, blob_ref, &size)) {
		report_pbms_error(pbcon);
		RETURN_FALSE;
	}
	
    RETURN_LONG(size);
}
/* }}} */

/*------------------------------------------------------*/
/*------------------------------------------------------*/
static zend_function_entry php_pbms_functions[] = {
	PHP_FE(pbms_errno, NULL)
	PHP_FE(pbms_error, NULL)
	PHP_FE(pbms_connect, NULL)
	PHP_FE(pbms_connection_pool_size, NULL)
	PHP_FE(pbms_pconnect, NULL)
	PHP_FE(pbms_close, NULL)
	PHP_FE(pbms_print_con_info, NULL)

	PHP_FE(pbms_add_metadata, NULL)
	PHP_FE(pbms_clear_metadata, NULL)
	PHP_FE(pbms_reset_metadata, NULL)
	PHP_FE(pbms_next_metadata, NULL)
	PHP_FE(pbms_get_metadata_value, NULL)

	PHP_FE(pbms_get_md5_digest, NULL)
	PHP_FE(pbms_get_info, NULL)

	PHP_FE(pbms_put_data, NULL)
	PHP_FE(pbms_get_data, NULL)
	PHP_FE(pbms_get_data_range, NULL)
	PHP_FE(pbms_read_stream, NULL)
	PHP_FE(pbms_write_stream, NULL)
	PHP_FE(pbms_write_stream_range, NULL)
	PHP_FE(pbms_is_blob_reference, NULL)
	PHP_FE(pbms_get_blob_size, NULL)

	{NULL, NULL, NULL}
};


#if ZEND_MODULE_API_NO >= 20050922
static zend_module_dep pbms_module_deps[] = {
    ZEND_MOD_REQUIRED("mysql")
    {NULL, NULL, NULL}
};
#endif

zend_module_entry pbms_module_entry = {
#if ZEND_MODULE_API_NO >= 20050922
	STANDARD_MODULE_HEADER_EX, NULL,
    pbms_module_deps,
#else
	STANDARD_MODULE_HEADER,
#endif
	"pbms",						/* PHP extension name (PBMS)*/
	php_pbms_functions,		    /* Functions*/
	ZEND_MINIT(pbms),	        /* MINIT*/
	ZEND_MSHUTDOWN(pbms),		/* MSHUTDOWN*/
	ZEND_RINIT(pbms),		    /* RINIT*/
	ZEND_RSHUTDOWN(pbms),	    /* RSHUTDOWN*/
	PHP_MINFO(pbms),			/* MINFO*/
#if ZEND_MODULE_API_NO >= 20010901
	"1.0",
#endif
	STANDARD_MODULE_PROPERTIES
};

#ifdef COMPILE_DL_PBMS
	ZEND_GET_MODULE(pbms)		/* for PHP extensions built as shared modules*/
#endif

#endif /* HAVE_PBMSLIB_H */
